require File.expand_path('../../spec_helper.rb', __FILE__)
include Sokoban

# Test input: One move away from level 1 completion
# Strip out whitespace due to multiline heredoc - ugly & effective
INPUT = <<-eos.gsub(/\s+/, "")
  WAAAWWWAWAASSWAASSSDDDDDDDDDDDDDSDWAAAAAAAAAWWWAWAASSSWWAASSSDDD
  DDDDDDDDDWDSASDWAAAAAAAAWWWAAWASSSWWAASSSDDDDDDDDDDDDSDWAAAAAAAA
  WWWAWAWWDSSWWAASSSSSWWAASSSDDDDDDDDDDDWDSASDWAAAAAAAWWWAAWASSSWW
  AASSSDDDDDDDDDDDDDAAAAAAAAAAAAAWAASDDDDDDDDDDDDD
eos

describe Level do
  it "defaults to level 1" do
    expect(Level.new.lnum).to eq(1)
  end

  it "sets lnum to the given level" do
    expect(Level.new(42).lnum).to eq(42)
  end

  it "initializes moves to 0" do
    expect(Level.new.moves).to eq(0)
  end

  it "initializes pushes to 0" do
    expect(Level.new.pushes).to eq(0)
  end

  it "starts level 1 with 6 free boxes" do
    expect(Level.new.free_boxes).to eq(6)
  end

  it "raises RangeError when level is out of range" do
    expect { Level.new(54) }.to raise_error(RangeError)
  end

  it "accepts valid moves" do
    level = Level.new
    level.parse_move(INPUT)
    expect(level.moves).to eq(INPUT.size)
  end

  it "ignores invalid moves" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("F") # shouldn"t increment moves
    expect(level.moves).to eq(INPUT.size)
  end

  it "restarts" do
    level = Level.new
    level.parse_move(INPUT)
    expect(level.free_boxes).to eq(1)
    level.parse_move("R") # restart
    expect(level.free_boxes).to eq(6)
  end

  it "plays level 1" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("D") # right
    expect(level.free_boxes).to eq(0) # Victory!
  end

  it "increments pushes on box move" do
    level = Level.new
    level.parse_move(INPUT)
    expect(level.pushes).to eq(96)
    level.parse_move("D") # right
    expect(level.pushes).to eq(97)
  end

  it "decrements pushes on box move undo" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("D") # right
    expect(level.pushes).to eq(97)
    level.parse_move("U") # undo
    expect(level.pushes).to eq(96)
  end

  it "ignores moves into walls" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("A") # left
    expect(level.moves).to eq(INPUT.size + 1)
    # Try to move into wall
    level.parse_move("W") # up
    expect(level.moves).to eq(INPUT.size + 1)
  end

  it "ignores moves into non-free boxes" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("SD")
    expect(level.moves).to eq(INPUT.size + 2)
    # Try to move into non-free box
    level.parse_move("D") # right
    expect(level.moves).to eq(INPUT.size + 2)
  end

  it "ignores pushes into walls" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("SDW") # down, right, up
    expect(level.pushes).to eq(97)
    # Try to push box into wall
    level.parse_move("W") # up
    expect(level.pushes).to eq(97)
  end

  it "ignores pushes into non-free boxes" do
    level = Level.new
    level.parse_move(INPUT)
    level.parse_move("WDS") # up, right, down
    expect(level.pushes).to eq(97)
    level.parse_move("AS") # left, down
    # Try to push box into non-free box
    level.parse_move("D") # right
    expect(level.pushes).to eq(97)
  end
end
